# STA 2102/450 Spring 2000 - S Functions for Assignment #3 - R. M. Neal


# NUMERICALLY INTEGRATE A FUNCTION USING THE MID-POINT RULE.  Takes as
# arguments the function to integrate, the low and high bounds for the
# integration, and the number of points to evaluate the function at.
# Any arguments beyond these are passed as additional arguments to the
# function being integrated.

midpoint.rule <- function (f, a, b, k, ...)
{
  h <- (b-a)/k
  s <- 0

  for (i in 1:k)
  { s <- s + f(a+(i-0.5)*h,...)
  }

  s * h
}


# THE UNORMALIZED PRIOR DENSITY FOR THETA IN THE +1/-1 MODEL.  The valid
# range for theta is (0,1), with the endpoints not included.

uprior <- function (theta)
{
  (1/sqrt(theta)) * (1/sqrt(1-theta)) * (0.2 + exp(-100*(theta-0.5)^2))
}


# FIND THE COUNT OF CHANGES IN THE DATA SEQUENCE.  Given a sequence of numbers,
# it returns the number of times the number changes from one time to the next.

change.count <- function (x)
{ 
  n <- length(x)

  sum (x[2:n]!=x[1:(n-1)])
}


# THE LIKELIHOOD FUNCTION.  Passed the number of data points and the number
# of changes from one data point to the next.  Does not leave out any constant 
# factors, so that it can be used when finding the Bayes factor.

likelihood <- function (theta, n, c)
{ 
  0.5 * theta^(n-1-c) * (1-theta)^c
}


# THE UNNORMALIZED POSTERIOR DENSITY.  Passed the total number of data points
# and count of the number of times the data changed from one time to the next.
# The unnormalized prior is used, and the likelihood with all factors included.

upost <- function (theta, n, c)
{
  uprior(theta) * likelihood(theta,n,c)
}


# COMPUTE THE NORMALIZING CONSTANT FOR THE PRIOR.  Passed the number of
# points to use for the integration.

prior.norm <- function (k)
{
  midpoint.rule (uprior, 0, 1, k)
}


# PREDICT THE NEXT DATA POINT.  Passed the previous data points, and the number
# of points to use for the integrations using the midpoint rule.  Integrates 
# the prediction for the next point given theta with respect to the unnormalized
# posterior distribution of theta.  Since the posterior is unnormalized, this 
# results in unnormalized probabilities of the next point being +1 and -1, 
# which are normalized to get the probability of the next point being +1.

predict <- function (x,k)
{
  n <- length(x)
  c <- change.count(x)

  upr.same <- midpoint.rule (
               function (theta,n,c) theta * upost(theta,n,c), 0, 1, k, n, c)
  upr.diff <- midpoint.rule (
               function (theta,n,c) (1-theta) * upost(theta,n,c), 0, 1, k, n, c)

  pr.same <- upr.same / (upr.same+upr.diff)
  pr.diff <- upr.diff / (upr.same+upr.diff)

  if (x[n]==+1)  
  { pr.same
  }
  else
  { pr.diff
  }
}


# ANOTHER FUNCTION TO PREDICT THE NEXT DATA POINT.  Find the normalized
# probabilities for the next point by first finding the normalizing constant
# for the prior and probability of the observed data.  Useful as a check on
# the function above.  (Actually, the prior normalizing constant cancels out,
# but I've computed it here anyway.)

predict2 <- function (x,k)
{
  n <- length(x)
  c <- change.count(x)
  
  pnorm <- prior.norm(k)
  pdata <- midpoint.rule (upost, 0, 1, k, n, c) / pnorm

  pr.same <- midpoint.rule (
               function (theta,n,c) theta * upost(theta,n,c) / pnorm, 
               0, 1, k, n, c) / pdata
  pr.diff <- midpoint.rule (
               function (theta,n,c) (1-theta) * upost(theta,n,c) / pnorm, 
               0, 1, k, n, c) / pdata

  if (x[n]==+1)  
  { pr.same
  }
  else
  { pr.diff
  }
}


# FIND THE BAYES FACTOR VERSUS THE TRIVIAL MODEL.  Passed the data points,
# and the number of points to use for the integrations using the midpoint rule.

bayes.factor <- function (x,k)
{
  n <- length(x)
  c <- change.count(x)

  (midpoint.rule (upost, 0, 1, k, n, c) / prior.norm(k)) / 0.5^n
}


# TESTS OF THE MIDPOINT RULE.

midpoint.rule(function (x) 5, 1, 3, 3)
midpoint.rule(function (x) 3*x-1, -1, 1, 2)
midpoint.rule(function (x) x^2, 0, 1, 2)
midpoint.rule(function (x) x^2, 0, 1, 200)
midpoint.rule(function (x) 1/sqrt(x), 0, 1, 100)
midpoint.rule(function (x) 1/sqrt(x), 0, 1, 400)
midpoint.rule(function (x) 1/sqrt(x), 0, 1, 1600)


# CALCULATIONS FOR THE BAYESIAN MODEL.

# Calculation of the normalization constant for the prior.

prior.norm(10000)
prior.norm(40000)

x <- c(+1,+1,-1,+1,+1,+1,-1,-1,-1,+1,+1)

# Prediction of the probability that next data point is +1.

predict(x,1000)
predict(x,2000)

# Prediction using the second function (should be similar to above).

predict2(x,1000)
predict2(x,2000)

# Calculation of the Bayes factor versus the trivial model.

bayes.factor(x,10000)
bayes.factor(x,40000)

